'''
CA: 이제 데이터 저장 기능을 추가한 버전을 설명드리겠습니다:
'''


import sys
import os
import socket
import json
import numpy as np
import pandas as pd
from datetime import datetime
from PyQt6.QtWidgets import (QApplication, QMainWindow, QPushButton, QFileDialog,
                           QVBoxLayout, QHBoxLayout, QWidget, QCheckBox, QLabel,
                           QSpinBox, QGroupBox)
from PyQt6.QtCore import QTimer
from pyqtgraph import PlotWidget

class FFTVisualizerWithSave(QMainWindow):
    def __init__(self, port=5000):
        super().__init__()
        self.setWindowTitle('FFT Visualizer with Data Saving')
        
        # 데이터 저장 관련 변수
        self.recording = False
        self.recorded_data = []
        self.auto_save_count = 0
        self.save_directory = ''
        
        self.setup_ui()
        self.setup_network(port)
        self.setup_timers()

    def setup_ui(self):
        # Create central widget and main layout
        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        self.main_layout = QVBoxLayout(central_widget)
        
        # Plot widget
        self.plot_widget = PlotWidget()
        self.main_layout.addWidget(self.plot_widget)
        
        # Set up the plot
        self.plot_widget.setBackground('w')
        self.plot_widget.setTitle("FFT Visualization")
        self.plot_widget.setLabel('left', 'Amplitude')
        self.plot_widget.setLabel('bottom', 'Frequency (Hz)')
        self.curve = self.plot_widget.plot(pen='b')
        
        # 데이터 저장 컨트롤 그룹 추가
        save_group = QGroupBox("Data Recording")
        save_layout = QVBoxLayout(save_group)
        
        # 수동 저장 버튼
        save_buttons_layout = QHBoxLayout()
        self.record_button = QPushButton("Start Recording")
        self.record_button.setCheckable(True)
        self.record_button.clicked.connect(self.toggle_recording)
        save_buttons_layout.addWidget(self.record_button)
        
        self.save_button = QPushButton("Save Recorded Data")
        self.save_button.clicked.connect(self.save_recorded_data)
        self.save_button.setEnabled(False)
        save_buttons_layout.addWidget(self.save_button)
        
        save_layout.addLayout(save_buttons_layout)
        
        # 자동 저장 설정
        auto_save_layout = QHBoxLayout()
        self.auto_save_cb = QCheckBox("Auto Save")
        self.auto_save_cb.stateChanged.connect(self.toggle_auto_save)
        auto_save_layout.addWidget(self.auto_save_cb)
        
        auto_save_layout.addWidget(QLabel("Save every"))
        self.auto_save_interval = QSpinBox()
        self.auto_save_interval.setRange(1, 1000)
        self.auto_save_interval.setValue(100)
        auto_save_layout.addWidget(self.auto_save_interval)
        auto_save_layout.addWidget(QLabel("samples"))
        
        save_layout.addLayout(auto_save_layout)
        
        # 저장 경로 설정
        path_layout = QHBoxLayout()
        self.path_button = QPushButton("Set Save Directory")
        self.path_button.clicked.connect(self.set_save_directory)
        path_layout.addWidget(self.path_button)
        
        self.path_label = QLabel("No directory selected")
        path_layout.addWidget(self.path_label)
        
        save_layout.addLayout(path_layout)
        
        # 메인 레이아웃에 저장 그룹 추가
        self.main_layout.addWidget(save_group)

    # def setup_ui(self):
    #     # ... 기존 UI 코드는 유지 ...
        
    #     # 데이터 저장 컨트롤 그룹 추가
    #     save_group = QGroupBox("Data Recording")
    #     save_layout = QVBoxLayout(save_group)
        
    #     # 수동 저장 버튼
    #     save_buttons_layout = QHBoxLayout()
    #     self.record_button = QPushButton("Start Recording")
    #     self.record_button.setCheckable(True)
    #     self.record_button.clicked.connect(self.toggle_recording)
    #     save_buttons_layout.addWidget(self.record_button)
        
    #     self.save_button = QPushButton("Save Recorded Data")
    #     self.save_button.clicked.connect(self.save_recorded_data)
    #     self.save_button.setEnabled(False)
    #     save_buttons_layout.addWidget(self.save_button)
        
    #     save_layout.addLayout(save_buttons_layout)
        
    #     # 자동 저장 설정
    #     auto_save_layout = QHBoxLayout()
    #     self.auto_save_cb = QCheckBox("Auto Save")
    #     self.auto_save_cb.stateChanged.connect(self.toggle_auto_save)
    #     auto_save_layout.addWidget(self.auto_save_cb)
        
    #     auto_save_layout.addWidget(QLabel("Save every"))
    #     self.auto_save_interval = QSpinBox()
    #     self.auto_save_interval.setRange(1, 1000)
    #     self.auto_save_interval.setValue(100)
    #     auto_save_layout.addWidget(self.auto_save_interval)
    #     auto_save_layout.addWidget(QLabel("samples"))
        
    #     save_layout.addLayout(auto_save_layout)
        
    #     # 저장 경로 설정
    #     path_layout = QHBoxLayout()
    #     self.path_button = QPushButton("Set Save Directory")
    #     self.path_button.clicked.connect(self.set_save_directory)
    #     path_layout.addWidget(self.path_button)
        
    #     self.path_label = QLabel("No directory selected")
    #     path_layout.addWidget(self.path_label)
        
    #     save_layout.addLayout(path_layout)
        
    #     # 메인 레이아웃에 저장 그룹 추가
    #     self.main_layout.addWidget(save_group)

    # def setup_network(self, port):
    #     self.port = port
    #     self.connected = False
        
    #     # Create socket
    #     self.client_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    #     self.client_socket.setblocking(False)
        
    #     # Start connection timer
    #     self.connect_timer = QTimer()
    #     self.connect_timer.timeout.connect(self.try_connect)
    #     self.connect_timer.start(1000)  # Try to connect every second

    # copied from 05_customizable_fft_spectrum_visualizer.py
    def setup_network(self, port):
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.bind(('localhost', port))
        self.server_socket.listen(1)
        self.server_socket.setblocking(False)
        
        self.client_socket = None
        self.connected = False

    # def setup_timers(self):
    #     # Update timer for plot
    #     self.update_timer = QTimer()
    #     self.update_timer.timeout.connect(self.update_plot)
    #     self.update_timer.start(50)  # Update every 50ms

    # copied from 05_customizable_fft_spectrum_visualizer.py
    def setup_timers(self):
        # 연결 타이머
        self.connect_timer = QTimer()
        self.connect_timer.timeout.connect(self.try_connect)
        self.connect_timer.start(100)
        
        # 업데이트 타이머
        self.update_timer = QTimer()
        self.update_timer.timeout.connect(self.update_plot)
        self.update_timer.start(50)

    # def try_connect(self):
    #     if not self.connected:
    #         try:
    #             self.client_socket.connect(('localhost', self.port))
    #             self.connected = True
    #             self.connect_timer.stop()
    #             self.statusBar().showMessage('Connected to server')
    #         except (ConnectionRefusedError, BlockingIOError):
    #             self.statusBar().showMessage('Waiting for server...')
    #         except Exception as e:
    #             self.statusBar().showMessage(f'Connection error: {e}')

    # copied from 05_customizable_fft_spectrum_visualizer.py
    def try_connect(self):
        if not self.connected:
            try:
                self.client_socket, _ = self.server_socket.accept()
                self.client_socket.setblocking(False)
                self.connected = True
                self.statusBar().showMessage('Connected!')
                self.connect_timer.stop()
            except BlockingIOError:
                pass
            except Exception as e:
                self.statusBar().showMessage(f'Connection error: {e}')

    def toggle_recording(self, checked):
        self.recording = checked
        if checked:
            self.record_button.setText("Stop Recording")
            self.recorded_data = []  # 새로운 기록 시작
            self.save_button.setEnabled(False)
        else:
            self.record_button.setText("Start Recording")
            self.save_button.setEnabled(True)

    def save_recorded_data(self):
        if not self.recorded_data:
            return
            
        file_name = f"fft_data_{datetime.now().strftime('%Y%m%d_%H%M%S')}.csv"
        if self.save_directory:
            file_path = os.path.join(self.save_directory, file_name)
        else:
            file_path = QFileDialog.getSaveFileName(
                self, 'Save Data', file_name, 'CSV files (*.csv)')[0]
            
        if file_path:
            try:
                # 데이터를 DataFrame으로 변환
                df = pd.DataFrame(self.recorded_data)
                # CSV 파일로 저장
                df.to_csv(file_path, index=False)
                self.statusBar().showMessage(f'Data saved to {file_path}')
            except Exception as e:
                self.statusBar().showMessage(f'Error saving data: {e}')

    def toggle_auto_save(self, state):
        if state and not self.save_directory:
            self.set_save_directory()
            if not self.save_directory:  # 디렉토리 선택이 취소된 경우
                self.auto_save_cb.setChecked(False)

    def set_save_directory(self):
        directory = QFileDialog.getExistingDirectory(
            self, "Select Save Directory")
        if directory:
            self.save_directory = directory
            self.path_label.setText(f"...{directory[-30:]}")
            self.statusBar().showMessage(f'Save directory set to {directory}')

    def update_plot(self):
        if not self.connected:
            return

        try:
            # data = self.client_socket.recv(4096).decode()

            k_buff_size = 4096
            data = b''
            while True:
                part = self.client_socket.recv(k_buff_size)
                data += part
                if len(part) < k_buff_size:
                    break
            data = data.decode()

            if data:
                fft_data = json.loads(data)
                frequencies = np.array(fft_data['frequencies'])
                amplitudes = np.array(fft_data['amplitudes'])
                
                # 데이터 기록
                if self.recording:
                    timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')
                    record = {
                        'timestamp': timestamp,
                        'frequencies': frequencies.tolist(),
                        'amplitudes': amplitudes.tolist()
                    }
                    self.recorded_data.append(record)
                
                # 자동 저장 처리
                if self.auto_save_cb.isChecked():
                    self.auto_save_count += 1
                    if self.auto_save_count >= self.auto_save_interval.value():
                        self.auto_save_data()
                        self.auto_save_count = 0
                
                # 플롯 업데이트
                self.curve.setData(frequencies, amplitudes)
                
        except BlockingIOError:
            pass
        except Exception as e:
            self.statusBar().showMessage(f'Error during update: {e}')
            self.connected = False
            self.connect_timer.start()

    def auto_save_data(self):
        if not self.recorded_data or not self.save_directory:
            return
            
        file_name = f"fft_auto_{datetime.now().strftime('%Y%m%d_%H%M%S')}.csv"
        file_path = os.path.join(self.save_directory, file_name)
        
        try:
            df = pd.DataFrame(self.recorded_data)
            df.to_csv(file_path, index=False)
            self.statusBar().showMessage(f'Auto-saved to {file_path}')
            self.recorded_data = []  # 저장 후 버퍼 클리어
        except Exception as e:
            self.statusBar().showMessage(f'Error auto-saving data: {e}')

    def closeEvent(self, event):
        # 프로그램 종료 시 마지막 데이터 저장
        if self.recorded_data and (self.recording or self.auto_save_cb.isChecked()):
            reply = QMessageBox.question(
                self, 'Save Data', 
                'Would you like to save the recorded data before closing?',
                QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes)
                
            if reply == QMessageBox.Yes:
                self.save_recorded_data()
        
        super().closeEvent(event)


def main():
    app = QApplication(sys.argv)
    app.setStyle('Fusion')
    window = FFTVisualizerWithSave()
    window.show()
    sys.exit(app.exec())

if __name__ == '__main__':
    main()

'''
이 버전의 주요 저장 기능을 설명드리겠습니다:

수동 저장 기능:

Record 버튼으로 데이터 기록 시작/중지
Save 버튼으로 기록된 데이터를 CSV 파일로 저장
저장 시 타임스탬프가 포함된 파일명 자동 생성


자동 저장 기능:

지정된 샘플 수마다 자동 저장
저장 디렉토리 지정 가능
자동 저장 간격 조절 가능


데이터 형식:

CSV 파일로 저장
타임스탬프, 주파수, 진폭 데이터 포함
pandas DataFrame 사용으로 데이터 처리 용이


사용자 편의 기능:

저장 경로 시각적 표시
상태 표시줄을 통한 저장 상태 알림
프로그램 종료 시 저장되지 않은 데이터 처리



이 구현을 사용하기 위해서는 추가로 pandas 패키지가 필요합니다:
pip install pandas

'''
